from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq

class TestDhanhq_ElectronicDelivery:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_generate_tpin_for_success(self, mock_read_request, dhanhq_obj):
        mock_read_request.return_value = { 'status': DhanHTTP.HttpResponseStatus.SUCCESS.value, 'remarks': '', 'data': '', }
        json_response = dhanhq_obj.generate_tpin()
        mock_read_request.assert_called_once_with('/edis/tpin')
        assert json_response['remarks'] == dhanhq.OTP_SENT # ToDo: Ideally, response.data should be set so
        assert json_response['data'] == ''

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_generate_tpin_for_failure(self, mock_read_request, dhanhq_obj):
        mock_read_request.return_value = {
            'status': DhanHTTP.HttpResponseStatus.FAILURE.value,
            'remarks': {
                    'error_code': 'HTPP500',
                    'error_type': 'Internal Server Error',
                    'error_message': 'HTPP500:Internal Server Error'
                },
            'data': '', }
        json_response = dhanhq_obj.generate_tpin()
        mock_read_request.assert_called_once_with('/edis/tpin')
        assert json_response['remarks'].startswith('status code :')
        assert json_response['data'] == ''

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    @patch("dhanhq.dhanhq._save_as_temp_html_file_and_open_in_browser")
    def test_open_browser_for_tpin_success(self, mock_save_and_open, mock_create_request, dhanhq_obj):
        # print(mock_create_request)
        # print(mock_save_and_open)
        mock_create_request.return_value = {
            'status': DhanHTTP.HttpResponseStatus.SUCCESS.value,
            'data': '{"edisFormHtml": "<html></html>"}'
        }
        response = dhanhq_obj.open_browser_for_tpin('isin', 1, 'exchange')
        assert response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        mock_save_and_open.assert_called_once()

    @patch('dhanhq.dhan_http.DhanHTTP.post')
    @patch("dhanhq.dhanhq._save_as_temp_html_file_and_open_in_browser")
    def test_open_browser_for_tpin_failure(self, mock_save_and_open, mock_create_request, dhanhq_obj):
        mock_create_request.return_value = {
            'status': DhanHTTP.HttpResponseStatus.FAILURE.value,
            'data': ''
        }
        response = dhanhq_obj.open_browser_for_tpin('isin', 1, 'exchange')
        assert response['status'] == DhanHTTP.HttpResponseStatus.FAILURE.value
        mock_save_and_open.assert_not_called()

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_edis_inquiry(self, mock_read_request, dhanhq_obj):
        isin = "123"
        endpoint = f'/edis/inquire/{isin}'
        dhanhq_obj.edis_inquiry(isin)
        mock_read_request.assert_called_once_with(endpoint)



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq


class TestDhanhq_ForeverOrder:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_forever(self,mock_read_request,dhanhq_obj):
        dhanhq_obj.get_forever()
        mock_read_request.assert_called_once_with('/forever/orders')

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_place_forever(self,mock_create_request, dhanhq_obj):
        endpoint = '/forever/orders'
        quantity = 100
        price = 108
        trigger_Price = 110
        dhanhq_obj.place_forever("security_id", "exchange_segment", "transaction_type",
                                 "product_type", "order_type", quantity, price, trigger_Price)
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.put")
    def test_modify_forever(self, mock_update_request, dhanhq_obj):
        order_id = 123
        endpoint = f'/forever/orders/{order_id}'
        quantity = 100
        price = 108
        trigger_price = 110
        disclosed_quantity = 555
        dhanhq_obj.modify_forever(order_id, "order_flag", "order_type", "leg_name",
                                  quantity, price, trigger_price, disclosed_quantity,"validity")
        mock_update_request.assert_called_once()
        assert mock_update_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.delete")
    def test_cancel_forever(self, mock_delete_request, dhanhq_obj):
        order_id = "123"
        endpoint = f'/forever/orders/{order_id}'
        dhanhq_obj.cancel_forever(order_id)
        mock_delete_request.assert_called_once_with(endpoint)




class Portfolio:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def get_holdings(self):
        """
        Retrieve all holdings bought/sold in previous trading sessions.

        Returns:
            dict: The response containing holdings data.
        """
        return self.dhan_http.get('/holdings')

    def get_positions(self):
        """
        Retrieve a list of all open positions for the day.

        Returns:
            dict: The response containing open positions.
        """
        return self.dhan_http.get('/positions')

    def convert_position(self, from_product_type, exchange_segment, position_type,
                         security_id, convert_qty, to_product_type):
        """
        Convert Position from Intraday to Delivery or vice versa.

        Args:
            from_product_type (str): The product type to convert from (e.g., CNC).
            exchange_segment (str): The exchange segment (e.g., NSE_EQ).
            position_type (str): The type of position (e.g., LONG).
            security_id (str): The ID of the security to convert.
            convert_qty (int): The quantity to convert.
            to_product_type (str): The product type to convert to (e.g., CNC).

        Returns:
            dict: The response containing the status of the conversion.
        """
        endpoint = '/positions/convert'
        payload = {
            "fromProductType": from_product_type,
            "exchangeSegment": exchange_segment,
            "positionType": position_type,
            "securityId": security_id,
            "convertQty": convert_qty,
            "toProductType": to_product_type
        }
        return self.dhan_http.post(endpoint, payload)






class Statement:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def get_trade_book(self, order_id=None):
        """
        Retrieve a list of all trades executed in a day.

        Args:
            order_id (str, optional): The ID of the specific order to retrieve trades for.

        Returns:
            dict: The response containing trade book data.
        """
        # ToDo: This is bad practice abusing REST principles.
        #  This should be broken into two different methods with appropriate REST convention-based URL.
        endpoint = f'/trades/{order_id if order_id is not None else ""}'
        return self.dhan_http.get(endpoint)

    def get_trade_history(self, from_date, to_date, page_number=0):
        """
        Retrieve the trade history for a specific date range.

        Args:
            from_date (str): The start date for the trade history.
            to_date (str): The end date for the trade history.
            page_number (int): The page number for pagination.

        Returns:
            dict: The response containing trade history data.
        """
        endpoint = f'/trades/{from_date}/{to_date}/{page_number}'
        return self.dhan_http.get(endpoint)

    def ledger_report(self, from_date, to_date):
        """
        Retrieve the ledger details for a specific date range.

        Args:
            from_date (str): The start date for the trade history.
            to_date (str): The end date for the trade history.

        Returns:
            dict: The response containing ledger details data.
        """
        endpoint = f'/ledger?from-date={from_date}&to-date={to_date}'
        return self.dhan_http.get(endpoint)



"""
This API gives you historical candle data for the desired scrip across segments & exchange.
This data is presented in the form of a candle and gives you timestamp, open, high, low, close (OHLC) & volume.
"""
import logging

class HistoricalData:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def intraday_minute_data(self, security_id, exchange_segment, instrument_type, from_date, to_date, interval=1):
        """
        Retrieve OHLC & Volume of minute candles for desired instrument for last 5 trading day.

        Args:
            security_id (str): The ID of the security.
            exchange_segment (str): The exchange segment (e.g., NSE, BSE).
            instrument_type (str): The type of instrument (e.g., stock, option).

        Returns:
            dict: The response containing intraday minute data.
        """
        if interval not in [1, 5, 15, 25, 60]:
            # Raising and catching an exception in same method is bad practice. Replaced it with this clean code
            err = "interval value must be ['1','5','15','25','60']"
            logging.error('Exception in dhanhq>>intraday_minute_data: %s', err)
            return {
                'status': 'failure',
                'remarks': err,
                'data': '',
            }
        endpoint = '/charts/intraday'
        payload = {
            'securityId': security_id,
            'exchangeSegment': exchange_segment,
            'instrument': instrument_type,
            'interval': interval,
            'fromDate': from_date,
            'toDate': to_date
        }
        return self.dhan_http.post(endpoint, payload)

    def historical_daily_data(self, security_id, exchange_segment, instrument_type, from_date, to_date, expiry_code=0):
        """
        Retrieve OHLC & Volume of daily candle for desired instrument.

        Args:
            security_id (str): Security ID of the instrument.
            exchange_segment (str): The exchange segment (e.g., NSE, BSE).
            instrument_type (str): The type of instrument (e.g., stock, option).
            expiry_code (str): The expiry code for derivatives.
            from_date (str): The start date for the historical data.
            to_date (str): The end date for the historical data.

        Returns:
            dict: The response containing historical daily data.
        """
        if expiry_code not in [0, 1, 2, 3]:
            err = "expiry_code value must be ['0','1','2','3']"
            logging.error('Exception in dhanhq>>historical_daily_data: %s', err)
            return {
                'status': 'failure',
                'remarks': err,
                'data': '',
            }
        endpoint = '/charts/historical'
        payload = {
            "securityId": security_id,
            "exchangeSegment": exchange_segment,
            "instrument": instrument_type,
            "expiryCode": expiry_code,
            "fromDate": from_date,
            "toDate": to_date
        }
        return self.dhan_http.post(endpoint, payload)



"""
This is a special file in Pytest that allows you to define fixtures that are available to all tests in your project,
without you having to import the fixture explicitly in your test files.
"""

import os
import pytest
from dotenv import load_dotenv

from dhanhq import DhanContext, dhanhq


# Load environment variables from .test.env file
load_dotenv('.test.env')

# Load Creds
API_CLIENT_ID = os.getenv('CLIENT_ID')
API_ACCESS_TOKEN = os.getenv('ACCESS_TOKEN')

@pytest.fixture
def api_client_id_fixture():
    return API_CLIENT_ID

@pytest.fixture
def api_access_token_fixture():
    return API_ACCESS_TOKEN

@pytest.fixture
def dhanhq_fixture(api_client_id_fixture, api_access_token_fixture):
    dhan_context = DhanContext(api_client_id_fixture, api_access_token_fixture)
    return dhanhq(dhan_context)

@pytest.fixture
def dhanhq_obj():
    dhan_context = DhanContext("test_client_id", "test_access_token")
    return dhanhq(dhan_context)




class TraderControl:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def kill_switch(self, action):
        """
        Control kill switch for user, which will disable trading for current trading day.

        Args:
            action (str): 'activate' or 'deactivate' to control the kill switch.

        Returns:
            dict: Status of Kill Switch for account.
        """
        action = action.upper()
        endpoint = f'/killswitch?killSwitchStatus={action}'
        # ToDo: This should have been an Update request aka HTTP-PUT and not HTTP-POST
        return self.dhan_http.post(endpoint)






class ForeverOrder:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def place_forever(self, security_id, exchange_segment, transaction_type, product_type, order_type,
                      quantity, price, trigger_Price, order_flag="SINGLE", disclosed_quantity=0, validity='DAY',
                      price1=0, trigger_Price1=0, quantity1=0, tag=None, symbol=""):
        """
        Place a new forever order in the Dhan account.

        Args:
            security_id (str): The ID of the security to trade.
            exchange_segment (str): The exchange segment (e.g., NSE, BSE).
            transaction_type (str): The type of transaction (BUY/SELL).
            product_type (str): The product type (e.g., CNC, INTRA).
            order_type (str): The type of order (LIMIT, MARKET, etc.).
            quantity (int): The quantity of the order.
            price (float): The price of the order.
            trigger_Price (float): The trigger price for the order.
            order_flag (str): The order flag (default is "SINGLE").
            disclosed_quantity (int): The disclosed quantity for the order.
            validity (str): The validity of the order (DAY, IOC, etc.).
            price1 (float): The secondary price for the order.
            trigger_Price1 (float): The secondary trigger price for the order.
            quantity1 (int): The secondary quantity for the order.
            tag (str): Optional correlation ID for tracking.
            symbol (str): The trading symbol for the order.

        Returns:
            dict: The response containing the status of the order placement.
        """
        endpoint = '/forever/orders'
        payload = {
            "orderFlag": order_flag,
            "transactionType": transaction_type.upper(),
            "exchangeSegment": exchange_segment.upper(),
            "productType": product_type.upper(),
            "orderType": order_type.upper(),
            "validity": validity.upper(),
            "tradingSymbol": symbol,
            "securityId": security_id,
            "quantity": int(quantity),
            "disclosedQuantity": int(disclosed_quantity),
            "price": float(price),
            "triggerPrice": float(trigger_Price),
            "price1": float(price1),
            "triggerPrice1": float(trigger_Price1),
            "quantity1": int(quantity1),
        }

        if tag not in (None, ''):
            payload["correlationId"] = tag

        return self.dhan_http.post(endpoint, payload)

    def modify_forever(self, order_id, order_flag, order_type, leg_name,
                       quantity, price, trigger_price, disclosed_quantity, validity):
        """
        Modify a forever order based on the specified leg name.
        The variables that can be modified include price, quantity, order type, and validity.

        Args:
            order_id (str): The ID of the order to modify.
            order_flag (str): The order flag indicating the type of order (e.g., SINGLE, OCO).
            order_type (str): The type of order (e.g., LIMIT, MARKET).
            leg_name (str): The name of the leg to modify.
            quantity (int): The new quantity for the order.
            price (float): The new price for the order.
            trigger_price (float): The trigger price for the order.
            disclosed_quantity (int): The disclosed quantity for the order.
            validity (str): The validity of the order.

        Returns:
            dict: The response containing the status of the modification.
        """
        endpoint = f'/forever/orders/{order_id}'
        payload = {
            "orderId": str(order_id),
            "orderFlag": order_flag,
            "orderType": order_type,
            "legName": leg_name,
            "quantity": quantity,
            "disclosedQuantity": disclosed_quantity,
            "price": price,
            "triggerPrice": trigger_price,
            "validity": validity
        }
        return self.dhan_http.put(endpoint, payload)

    def get_forever(self):
        """Retrieve a list of all existing Forever Orders."""
        endpoint = '/forever/orders'
        return self.dhan_http.get(endpoint)

    def cancel_forever(self, order_id):
        """Delete Forever orders using the order id of an order."""
        endpoint = f'/forever/orders/{order_id}'
        return self.dhan_http.delete(endpoint)



"""
    A module that has a class encapsulating connection context and connection mediums to Dhan APIs.

    This library provides methods to manage orders, retrieve market data,
    and perform various trading operations through the DhanHQ API.

    :copyright: (c) 2025 by Dhan.
    :license: see LICENSE for details.
"""

import logging

from dhanhq.dhan_http import DhanHTTP

class DhanContext:
    """
        A class that encapsulates connection context to Dhan APIs like client-id, access-tocken, base-url
        and passes this to all the connection protocols like http and websocket that it is composed of.
    """

    def __init__(self, client_id, access_token, disable_ssl=False, pool=None):
        try:
            self.client_id = client_id
            self.access_token = access_token
            self.dhan_http = DhanHTTP(client_id, access_token, disable_ssl, pool)

        except Exception as e:
            logging.error('Exception in dhanhq>>init : %s', e)

    def get_client_id(self):
        """
        Return client's id that is used to identify the client interacting with Dhan API
        Returns client_id that is used to identify the client interacting with Dhan API
        """
        return self.client_id

    def get_access_token(self):
        """
        Return authorization token that is used for connecting to Dhan API
        Returns access_token that is used for authorization in accessing Dhan API
        """
        return self.access_token

    def get_dhan_http(self):
        """
        Return HTTP Connection Request object that has all necessary context to connect to Dhan API

        Returns
        http_connection_request (DhanHTTP): DhanContext enabled HTTP Connection Request object
        """
        return self.dhan_http



import json
import os
from pprint import pprint

import pytest
from dhanhq import dhanhq

class TestE2E_Dhanhq_Portfolio:
    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get-current-holdings.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_holdings(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_holdings()
        pprint("actual_response = " + actual_response)
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_positions.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_positions(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_positions()
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/convert_position.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_convert_position(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.convert_position(dhanhq.CNC, dhanhq.NSE, "LONG", "security_id", -99, dhanhq.INTRA)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

class TestE2E_Dhanhq_Funds:
    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_fund_limits.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_fund_limits(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_fund_limits()
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/margin_calculator.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_margin_calculator(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.margin_calculator("security_id", dhanhq.NSE, dhanhq.BUY, 100, dhanhq.CNC, 99.99)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq

class TestDhanhq_Portfolio:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_positions(self, mock_read_request, dhanhq_obj):
        dhanhq_obj.get_positions()
        mock_read_request.assert_called_once_with('/positions')

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_holdings(self, mock_read_request, dhanhq_obj):
        dhanhq_obj.get_holdings()
        mock_read_request.assert_called_once_with('/holdings')

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_convert_position(self, mock_create_request, dhanhq_obj):
        dhanhq_obj.convert_position("from_product_type", "exchange_segment", "position_type", "security_id", "convert_qty", "to_product_type")
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == '/positions/convert'



"""
    The FullDepth class is designed to facilitate asynchronous communication with the DhanHQ API via WebSocket.
    It enables users to subscribe to market data for a list of instruments and receive real-time updates.

    :copyright: (c) 2025 by Dhan.
    :license: see LICENSE for details.
"""

import websockets
import asyncio
import struct
from datetime import datetime
import json


class FullDepth:
    # Constants
    """WebSocket URL for DhanHQ Live Market Feed"""
    depth_feed_wss = 'wss://depth-api-feed.dhan.co/twentydepth'

    """Constants for Exchange Segment"""
    NSE = 1
    NSE_FNO = 2

    """Constants for Request Code"""
    Depth_20 = 23

    def __init__(self, dhan_context, instruments):
        """Initializes the FullDepth instance with user credentials, instruments to subscribe, and callback functions."""
        self.client_id = dhan_context.get_client_id()
        self.access_token = dhan_context.get_access_token()
        self.instruments = instruments
        self.data = ""
        self._is_first_connect = True
        self.ws = None
        self.on_ticks = None
        self.loop = asyncio.get_event_loop()
        self.on_close = False

    def run_forever(self):
        """Starts the WebSocket connection and runs the event loop."""
        self.loop.run_until_complete(self.connect())

    def get_data(self):
        """Fetch instruments data while the event loop is open."""
        response = self.loop.run_until_complete(self.ws.recv())
        remaining_data = response
        bid_data = None
        ask_data = None
        
        while remaining_data:
            update = self.process_data(remaining_data)
            if not update:
                break
            
            # Store remaining data for next iteration
            remaining_data = update.pop("remaining_data", None)
            
            # Store bid/ask data separately
            if update['type'] == "Bid":
                bid_data = update
            else:
                ask_data = update
            
            # If we have both bid and ask data, format and print them
            if bid_data and ask_data and bid_data['security_id'] == ask_data['security_id']:
                formatted_data = self.combine_and_format_depth(bid_data, ask_data)
                print(f"\nExchange Segment: {formatted_data['exchange_segment']}, Security ID: {formatted_data['security_id']}")
                for line in formatted_data['depth']:
                    print(line)
                # Reset bid/ask data for this security ID
                bid_data = None
                ask_data = None
        
        return None

    def close_connection(self):
        """Close WebSocket connection with this."""
        return self.loop.run_until_complete(self.disconnect())

    async def connect(self):
        """Initiates the connection to the Websockets."""
        if not self.ws or self.ws.state == websockets.protocol.State.CLOSED:
            url = f"{FullDepth.depth_feed_wss}?token={self.access_token}&clientId={self.client_id}&authType=2"
            self.ws = await websockets.connect(url)
            await self.subscribe_instruments()
        else:
            try:
                await self.ws.ping()
            except websockets.ConnectionClosed:
                self.ws = None
                await self.connect()

    async def get_instrument_data(self):
        """Fetches data and processes messages one at a time"""
        response = await self.ws.recv()
        remaining_data = response
        
        while remaining_data:
            update = self.process_data(remaining_data)
            if not update:
                break
                
            # Store remaining data for next iteration
            remaining_data = update.pop("remaining_data", None)
            yield update

    async def disconnect(self):
        """Closes the WebSocket connection and sends a disconnect message."""
        if self.ws:
            disconnect_message = {
                "RequestCode": 12
            }
            await self.ws.send(json.dumps(disconnect_message))
            header_message = self.create_header(feed_request_code=12, message_length=83, client_id=self.client_id)
            await self.ws.send(header_message)
        print("Connection closed!")

    """Creating Instruments List to be subscribed"""
    def validate_and_process_tuples(self, tuples_list, batch_size=50):
        """Create a list of all instruments to be added and a maximum of 50 instruments in single socket"""
        # Convert 2-element tuples to standard format
        processed_tuples = []
        for tup in tuples_list:
            if len(tup) == 2:
                exchange, token = tup
                processed_tuples.append((exchange, token))
            else:
                exchange, token, _ = tup  # Ignore the third element if present
                processed_tuples.append((exchange, token))

        # Eliminate duplicates
        processed_tuples = list(set(processed_tuples))
        
        # Create batches of 50 instruments
        return [processed_tuples[i:i+batch_size] for i in range(0, len(processed_tuples), batch_size)]

    async def subscribe_instruments(self):
        """Subscribe Instruments on the Open Websocket"""
        instrument_batches = self.validate_and_process_tuples(self.instruments)
        for batch in instrument_batches:
            subscription_message = {
                "RequestCode": self.Depth_20,
                "InstrumentCount": len(batch),
                "InstrumentList": [
                    {
                        "ExchangeSegment": self.get_exchange_segment(ex),
                        "SecurityId": token
                    } for ex, token in batch
                ]
            }
            await self.ws.send(json.dumps(subscription_message))
            print(subscription_message)

    def get_exchange_segment(self, exchange_code):
        """Convert numeric exchange code to string representation"""
        exchange_map = {
            1: "NSE_EQ",
            2: "NSE_FNO"
        }
        return exchange_map.get(exchange_code, str(exchange_code))

    def process_data(self, data):
        """Read binary data and process messages one at a time"""
        offset = 0
        
        # Process just one message and return it
        if offset + 12 <= len(data):  # Need 12 bytes for complete header
            header = struct.unpack('<hBBiI', data[offset:offset+12])
            msg_length = header[0]      # Message Length (2 bytes)
            msg_code = header[1]        # Feed Response Code (1 byte)
            exchange_segment = header[2] # Exchange Segment (1 byte)
            security_id = header[3]      # Security ID (4 bytes)
            
            if msg_length > 0 and offset + msg_length <= len(data):
                message_data = data[offset:offset+msg_length]
                if msg_code in [41, 51]:  # Either Bid (41) or Ask (51) data
                    depth_data = self.process_depth_data(message_data, is_bid=(msg_code == 41))
                    return {
                        "exchange_segment": exchange_segment,
                        "security_id": security_id,
                        "type": "Bid" if msg_code == 41 else "Ask",
                        "depth": depth_data["depth"],
                        "remaining_data": data[offset+msg_length:] if offset+msg_length < len(data) else None
                    }
                elif msg_code == 50:
                    self.server_disconnection(message_data)
        
        return None

    def process_depth_data(self, data, is_bid=True):
        """Parse and process depth data for both bid and ask"""
        # Only unpack the header first (12 bytes)
        header = struct.unpack('<hBBiI', data[0:12])
        exchange_segment = header[2]
        sec_id = header[3]
        
        # Skip the header and process the body directly
        market_depth_binary = data[12:]
        packet_format = '<dII'  # float64 (price) + uint32 (quantity) + uint32 (orders)
        packet_size = struct.calcsize(packet_format)
        depth = []
        
        for i in range(20):
            start_idx = i * packet_size
            end_idx = start_idx + packet_size
            
            if end_idx > len(market_depth_binary):
                break
                
            current_packet = struct.unpack(packet_format, market_depth_binary[start_idx:end_idx])
            depth.append({
                "price": current_packet[0],
                "quantity": current_packet[1],
                "orders": current_packet[2]
            })
        
        return {
            "type": 'Bid Data' if is_bid else 'Ask Data',
            "exchange_segment": exchange_segment,
            "security_id": sec_id,
            "depth": depth
        }

    def combine_and_format_depth(self, bid_data, ask_data):
        """Combine and sort bid/ask data, then format for display"""
        # Sort bid prices in descending order (highest first)
        bid_depth = sorted(bid_data["depth"], key=lambda x: x["price"], reverse=True)
        # Sort ask prices in ascending order (lowest first)
        ask_depth = sorted(ask_data["depth"], key=lambda x: x["price"])
        
        formatted_output = []
        for bid, ask in zip(bid_depth, ask_depth):
            line = (
                f"bid : {{price:{bid['price']:.2f}, quantity:{bid['quantity']}, no_of_orders:{bid['orders']}}} | "
                f"ask : {{price:{ask['price']:.2f}, quantity:{ask['quantity']}, no_of_orders:{ask['orders']}}}"
            )
            formatted_output.append(line)
        
        return {
            "exchange_segment": bid_data["exchange_segment"],
            "security_id": bid_data["security_id"],
            "depth": formatted_output
        }

    def server_disconnection(self, data):
        """Parse and process server disconnection error"""
        disconnection_packet = [struct.unpack('<hBBiI', data[0:12])]
        self.on_close = False
        if disconnection_packet[0][5] == 805:
            print ("Disconnected: No. of active websocket connections exceeded")
            self.on_close = True
        elif disconnection_packet[0][5] == 806:
            print ("Disconnected: Subscribe to Data APIs to continue")
            self.on_close = True
        elif disconnection_packet[0][5] == 807:
            print ("Disconnected: Access Token is expired")
            self.on_close = True
        elif disconnection_packet[0][5] == 808:
            print ("Disconnected: Invalid Client ID")
            self.on_close = True
        elif disconnection_packet[0][5] == 809:
            print ("Disconnected: Authentication Failed - check ")
            self.on_close = True

    async def on_connection_opened(self, websocket):
        "Callback function executed when the WebSocket connection is opened."
        await websocket.send(self.create_subscription_packet(self.instruments, subscribe=True))

        while True:
            response = await websocket.recv()
            await self.on_message_received(response)

    def pad_with_zeros(self, data, length):
        """Pads a binary data string with zeros to a specified length for server to read."""
        data = data.ljust(length, b'\0')
        return data

    def create_header(self, feed_request_code, message_length, client_id):
        """Creates header packet for the subscription packet."""
        dhan_auth = b"\0" * 50
        header = struct.pack('<bH30s50s', feed_request_code, message_length, client_id.encode('utf-8'), dhan_auth)
        return header

    def utc_time(self, epoch_time):
        """Converts EPOCH time to UTC time."""
        return datetime.utcfromtimestamp(epoch_time).strftime('%H:%M:%S')

    def create_subscription_packet(self, instruments, feed_request_code):
        """Creates the subscription packet with specified instruments and subscription code"""
        num_instruments = len(instruments)

        header = self.create_header(feed_request_code = feed_request_code,
                                    message_length=83 + 4 + num_instruments * 21,
                                    client_id=self.client_id)
        num_instruments_bytes = struct.pack('<I', num_instruments)
        instrument_info = b""
        for exchange_segment, security_id in instruments:
            instrument_info += struct.pack('<B20s', exchange_segment, security_id.encode('utf-8'))

        instruments = [(0, "")]
        for i in range(50 - num_instruments):
            instrument_info += struct.pack('<B20s', instruments[0][0], instruments[0][1].encode('utf-8'))

        subscription_packet = header + num_instruments_bytes + instrument_info
        return subscription_packet

    def subscribe_symbols(self, symbols):
        """Function to subscribe to additional symbols when connection is already established."""
        # Update the instruments list
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.update(symbols)
        self.instruments = list(unique_symbols_set)

        # If the WebSocket is open, send the subscription packet for the new symbols
        if self.ws and not self.ws.closed:
            # Prepare the instruments list for subscription
            new_instrument_list = self.validate_and_process_tuples(symbols)
            for instrument_type, instrument_groups in new_instrument_list.items():
                for instrument_group in instrument_groups:
                    for i in range(0, len(instrument_group), 50):
                        batch = instrument_group[i:i+50]
                        subscription_message = {
                            "RequestCode": int(instrument_type),
                            "InstrumentCount": len(batch),
                            "InstrumentList": [
                                {
                                    "ExchangeSegment": self.get_exchange_segment(ex),
                                    "SecurityId": token
                                } for ex, token in batch
                            ]
                        }
                        asyncio.ensure_future(self.ws.send(json.dumps(subscription_message)))

    def unsubscribe_symbols(self, symbols):
        """Function to unsubscribe symbols from connection when connection is already active."""
        # Update the instruments list by removing the specified symbols
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.difference_update(symbols)
        self.instruments = list(unique_symbols_set)

        # If the WebSocket is open, send the unsubscription packet for the symbols
        if self.ws and not self.ws.closed:
            instrument_list_to_unsubscribe = self.validate_and_process_tuples(symbols)
            for instrument_type, instrument_groups in instrument_list_to_unsubscribe.items():
                for instrument_group in instrument_groups:
                    for i in range(0, len(instrument_group), 50):
                        batch = instrument_group[i:i+50]
                        unsubscription_message = {
                            "RequestCode": int(instrument_type) + 1,
                            "InstrumentCount": len(batch),
                            "InstrumentList": [
                                {
                                    "ExchangeSegment": self.get_exchange_segment(ex),
                                    "SecurityId": token
                                } for ex, token in batch
                            ]
                        }
                        asyncio.ensure_future(self.ws.send(json.dumps(unsubscription_message)))


"""
    A class that has core DhanHQ APIs.

    This library provides methods to manage orders, retrieve market data,
    and perform various trading operations through the DhanHQ API.

    :copyright: (c) 2025 by Dhan.
    :license: see LICENSE for details.
"""

from datetime import datetime, timedelta, timezone


from dhanhq import (Order, ForeverOrder, Portfolio, Statement, TraderControl, Security,
                    HistoricalData, OptionChain, MarketFeed, Funds)


class dhanhq(Order, ForeverOrder, Portfolio, Funds, Statement, TraderControl, Security,
             MarketFeed, HistoricalData, OptionChain):
    """DhanHQ Class having Core APIs"""

    """Constants for Exchange Segment"""
    NSE = 'NSE_EQ'
    BSE = 'BSE_EQ'
    CUR = 'NSE_CURRENCY'
    MCX = 'MCX_COMM'
    FNO = 'NSE_FNO'
    NSE_FNO = 'NSE_FNO'
    BSE_FNO = 'BSE_FNO'
    INDEX = 'IDX_I'

    """Constants for Transaction Type"""
    BUY = 'BUY'
    SELL = 'SELL'

    """Constants for Product Type"""
    CNC = 'CNC'
    INTRA = "INTRADAY"
    MARGIN = 'MARGIN'
    CO = 'CO'
    BO = 'BO'
    MTF = 'MTF'

    """Constants for Order Type"""
    LIMIT = 'LIMIT'
    MARKET = 'MARKET'
    SL = "STOP_LOSS"
    SLM = "STOP_LOSS_MARKET"

    """Constants for Validity"""
    DAY = 'DAY'
    IOC = 'IOC'


    def __init__(self, dhan_context):
        for parent in [Order, ForeverOrder, Portfolio, Funds, Statement, TraderControl, Security,
                       MarketFeed, HistoricalData, OptionChain]:
            parent.__init__(self,dhan_context)
        self.dhan_http = dhan_context.get_dhan_http()

    @staticmethod
    def convert_to_date_time(self, epoch):
        """
        Convert EPOCH time to Python datetime object in IST.

        Args:
            epoch (int): The EPOCH time to convert.

        Returns:
            datetime: Corresponding datetime object in IST.
        """
        IST = timezone(timedelta(hours=5, minutes=30))
        dt = datetime.fromtimestamp(epoch, IST)

        if dt.time() == datetime.min.time():
            return dt.date()
        return dt



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq


class TestDhanhq_TraderControls:
    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_kill_switch(self, mock_post_request, dhanhq_obj):
        action = "action"
        endpoint = f'/killswitch?killSwitchStatus={action.upper()}'
        dhanhq_obj.kill_switch(action)
        mock_post_request.assert_called_once_with(endpoint)



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq


class TestDhanhq_Funds:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_fund_limits(self, mock_read_request, dhanhq_obj):
        endpoint = '/fundlimit'
        dhanhq_obj.get_fund_limits()
        mock_read_request.assert_called_once_with(endpoint)

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_margin_calculator(self, mock_create_request, dhanhq_obj):
        endpoint = '/margincalculator'
        quantity=100
        price = 99.99
        dhanhq_obj.margin_calculator("security_id", "exchange_segment", "transaction_type",
                                     100, "product_type", price)
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == endpoint




class Funds:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def get_fund_limits(self):
        """
        Get all information of your trading account like balance, margin utilized, collateral, etc.

        Returns:
            dict: The response containing fund limits data.
        """
        endpoint = '/fundlimit'
        return self.dhan_http.get(endpoint)

    def margin_calculator(self, security_id, exchange_segment, transaction_type, quantity,
                          product_type, price, trigger_price=0):
        """
        Calculate the margin required for a trade based on the provided parameters.

        Args:
            security_id (str): The ID of the security for which the margin is to be calculated.
            exchange_segment (str): The exchange segment (e.g., NSE_EQ) where the trade will be executed.
            transaction_type (str): The type of transaction (BUY/SELL).
            quantity (int): The quantity of the security to be traded.
            product_type (str): The product type (e.g., CNC, INTRA) of the trade.
            price (float): The price at which the trade will be executed.
            trigger_price (float, optional): The trigger price for the trade. Defaults to 0.

        Returns:
            dict: The response containing the margin calculation result.
        """
        endpoint = '/margincalculator'
        payload = {
            "securityId": security_id,
            "exchangeSegment": exchange_segment.upper(),
            "transactionType": transaction_type.upper(),
            "quantity": int(quantity),
            "productType": product_type.upper(),
            "price": float(price)
        }
        # ToDo: Shouldn't price and trigger_price being float vlaues be rounded to 2 or 3 decimal places as precision??
        if trigger_price >= 0:
            payload["triggerPrice"] = float(trigger_price)

        return self.dhan_http.post(endpoint, payload)




import json
import os
from pprint import pprint

import pytest
from dhanhq import dhanhq

# from tests.conftest import api_access_token_fixture

class TestE2E_Dhanhq_Forever_Orders:
    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get-forever-orders-list.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_order_list(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_forever()
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/modify_pending_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_modify_order(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.modify_forever("order_id", "SINGLE", dhanhq.LIMIT, "TARGET_LEG",
                                                        100, 99.99, 99.99, 100,dhanhq.DAY)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/cancel_given_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_cancel_forever(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.cancel_forever("string")
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/place_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_place_forever(self, expected_dict, dhanhq_fixture):

        actual_response = dhanhq_fixture.place_forever("security_id", dhanhq.NSE, dhanhq.BUY, dhanhq.CNC,
                                                       dhanhq.LIMIT,100, 99.99, 99.99)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']



import json
from unittest.mock import patch

import pytest
import requests

from dhanhq.dhan_http import DhanHTTP

@pytest.fixture
def dhan_http():
    return DhanHTTP("test_client_id", "test_access_token")

@pytest.fixture
def mock_success_response():
    mock_response = requests.Response()
    mock_response.status_code = 200
    mock_response._content = b'{"data": "some data"}'
    return mock_response

@pytest.fixture
def mock_failure_response():
    mock_response = requests.Response()
    mock_response.status_code = 400
    mock_response._content = b'{"errorType": "test_error", "errorMessage": "test message"}'
    return mock_response

class TestDhan_Constructor:
    def test_init_success(self, dhan_http):
        application_json = "application/json"
        client_id = "test_client_id"
        access_token = "test_access_token"
        assert dhan_http.client_id == client_id
        assert dhan_http.access_token == access_token
        assert dhan_http.base_url == "https://api.dhan.co/v2"
        assert dhan_http.timeout == DhanHTTP.HTTP_DEFAULT_TIME_OUT
        assert dhan_http.header["access-token"] == access_token
        assert dhan_http.header["client-id"] == client_id
        assert dhan_http.header["Content-type"] == application_json
        assert dhan_http.header["Accept"] == application_json

class TestDhan_Private_ParseResponse:
    def test_parse_response_success_with_status_code_200(self, dhan_http, mock_success_response):
        json_response = dhan_http._parse_response(mock_success_response)
        assert json_response["status"] == "success"
        assert json_response["remarks"] == ""
        assert json_response["data"] == {"data": "some data"}

    def test_parse_response_success_with_status_code_299(self, dhan_http, mock_success_response):
        mock_success_response.status_code = 299
        json_response = dhan_http._parse_response(mock_success_response)
        assert json_response["status"] == "success"
        assert json_response["remarks"] == ""
        assert json_response["data"] == {"data": "some data"}

    def test_parse_response_error_with_status_code_400(self, dhan_http, mock_failure_response):
        json_response = dhan_http._parse_response(mock_failure_response)
        assert json_response["status"] == "failure"
        assert json_response["remarks"]["error_type"] == "test_error"
        assert json_response["remarks"]["error_message"] == "test message"
        assert json_response["data"] == ""

class TestDhan_Private_SendRequest:
    @patch("requests.Session.post")
    def test_send_request_with_payload_add_clientid_to_it(self,mock_requests_session_post, dhan_http):
        """Tests POST request with a payload and verifies payload content."""
        endpoint = "/endpoint"
        payload = {"key": "value"}
        dhan_http._send_request(DhanHTTP.HttpMethods.POST, endpoint, payload)
        mock_requests_session_post.assert_called_once_with(
            dhan_http.base_url + endpoint,
            data=json.dumps({**payload, "dhanClientId": dhan_http.client_id}),
            headers=dhan_http.header,
            timeout=dhan_http.timeout,
        )

    @patch("requests.Session.get")
    def test_send_request_get_no_payload(self, mock_requests_session_get, dhan_http):
        """Tests GET request with no payload."""
        endpoint = "/endpoint"
        dhan_http._send_request(DhanHTTP.HttpMethods.GET, endpoint)
        mock_requests_session_get.assert_called_once_with(
            dhan_http.base_url + endpoint,
            data=None,
            headers=dhan_http.header,
            timeout=dhan_http.timeout
        )

    @patch("requests.Session.post")
    def test_send_request_exception(self, mock_requests_session_post, dhan_http):
        """Tests exception handling in _send_request."""
        endpoint = "/endpoint"
        mock_requests_session_post.side_effect = requests.exceptions.ConnectionError("Test Error")
        response = dhan_http._send_request(DhanHTTP.HttpMethods.POST, endpoint, {"test": "test"})
        assert response['status'] == 'failure'

class TestDhan_CRUD_Methods:
    @patch("dhanhq.DhanHTTP._send_request")
    def test_get(self,mock_send_request,dhan_http):
        endpoint = "/endpoint"
        dhan_http.get(endpoint)
        mock_send_request.assert_called_once_with(DhanHTTP.HttpMethods.GET,endpoint,)

    @patch("dhanhq.DhanHTTP._send_request")
    def test_delete(self,mock_send_request,dhan_http):
        endpoint = "/endpoint"
        dhan_http.delete(endpoint)
        mock_send_request.assert_called_once_with(DhanHTTP.HttpMethods.DELETE, endpoint)

    @patch("dhanhq.DhanHTTP._send_request")
    def test_put(self,mock_send_request,dhan_http):
        endpoint = "/endpoint"
        payload = {"key": "value"}
        dhan_http.put(endpoint,payload)
        mock_send_request.assert_called_once_with(DhanHTTP.HttpMethods.PUT,endpoint,payload)

    @patch("dhanhq.DhanHTTP._send_request")
    def test_post(self,mock_send_request,dhan_http):
        endpoint = "/endpoint"
        payload = {"key": "value"}
        dhan_http.post(endpoint,payload)
        mock_send_request.assert_called_once_with(DhanHTTP.HttpMethods.POST, endpoint, payload)



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq


class TestDhanhq_Statement:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_trade_book_without_orderid(self, mock_read_request, dhanhq_obj):
        endpoint = '/trades/'
        dhanhq_obj.get_trade_book()
        mock_read_request.assert_called_once_with(endpoint)

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_trade_book_with_orderid(self, mock_read_request, dhanhq_obj):
        order_id = "order_id"
        endpoint = f'/trades/{order_id}'
        dhanhq_obj.get_trade_book(order_id)
        mock_read_request.assert_called_once_with(endpoint)

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_trade_history(self, mock_read_request, dhanhq_obj):
        from_date = "from_date"
        to_date = "to_date"
        page_number = "page_number"
        endpoint = f'/trades/{from_date}/{to_date}/{page_number}'
        dhanhq_obj.get_trade_history(from_date,to_date,page_number)
        mock_read_request.assert_called_once_with(endpoint)

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_ledger_report(self, mock_read_request, dhanhq_obj):
        from_date = "from_date"
        to_date = "to_date"
        endpoint = f'/ledger?from-date={from_date}&to-date={to_date}'
        dhanhq_obj.ledger_report(from_date,to_date)
        mock_read_request.assert_called_once_with(endpoint)

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_intraday_minute_data(self, mock_create_request, dhanhq_obj):
        endpoint = f'/charts/intraday'
        security_id="security_id"
        exchange_segment="exchange_segment"
        instrument_type="instrument_type"
        from_date="from_date"
        to_date="to_date"
        mock_create_request.return_value = { 'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.intraday_minute_data(security_id, exchange_segment, instrument_type, from_date, to_date)
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_intraday_minute_data_fails_for_bad_interval_input(self, mock_create_request, dhanhq_obj):
        security_id="security_id"
        exchange_segment="exchange_segment"
        instrument_type="instrument_type"
        from_date="from_date"
        to_date="to_date"
        interval=100
        json_response = dhanhq_obj.intraday_minute_data(security_id, exchange_segment, instrument_type, from_date, to_date, interval)
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.FAILURE.value
        mock_create_request.assert_not_called()

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_historical_daily_data(self, mock_create_request, dhanhq_obj):
        endpoint = f'/charts/historical'
        security_id='security_id'
        exchange_segment='exchange_segment'
        instrument_type='instrument_type'
        from_date="from_date"
        to_date="to_date"
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.historical_daily_data(security_id, exchange_segment, instrument_type, from_date, to_date)
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_historical_daily_data_fails_for_bad_expiry_code(self, mock_create_request, dhanhq_obj):
        security_id='security_id'
        exchange_segment='exchange_segment'
        instrument_type='instrument_type'
        from_date="from_date"
        to_date="to_date"
        bad_expiry_code=99
        json_response = dhanhq_obj.historical_daily_data(security_id, exchange_segment, instrument_type, from_date, to_date,bad_expiry_code)
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.FAILURE.value
        mock_create_request.assert_not_called()

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_ticker_data(self, mock_create_request, dhanhq_obj):
        endpoint = '/marketfeed/ltp'
        securities = {
            'exchange_segment1': 'security_id1',
            'exchange_segment2': 'security_id2'
        }
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.ticker_data(securities)
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint
        assert mock_create_request.call_args[0][1] == securities

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_ohlc_data(self, mock_create_request, dhanhq_obj):
        endpoint = '/marketfeed/ohlc'
        securities = {
            'exchange_segment1': 'security_id1',
            'exchange_segment2': 'security_id2'
        }
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.ohlc_data(securities)
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint
        assert mock_create_request.call_args[0][1] == securities

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_quote_data(self, mock_create_request, dhanhq_obj):
        endpoint = '/marketfeed/quote'
        securities = {
            'exchange_segment1': 'security_id1',
            'exchange_segment2': 'security_id2'
        }
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.quote_data(securities)
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint
        assert mock_create_request.call_args[0][1] == securities

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_option_chain(self, mock_create_request, dhanhq_obj):
        endpoint = '/optionchain'
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.option_chain("under_security_id", "under_exchange_segment", "expiry")
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_expiry_list(self, mock_create_request, dhanhq_obj):
        endpoint = '/optionchain/expirylist'
        mock_create_request.return_value = {'status': DhanHTTP.HttpResponseStatus.SUCCESS.value}
        json_response = dhanhq_obj.expiry_list("under_security_id", "under_exchange_segment")
        mock_create_request.assert_called_once()
        assert json_response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value
        assert mock_create_request.call_args[0][0] == endpoint



from json import loads as json_loads
from pathlib import Path
from webbrowser import open as web_open
import logging
import requests

from dhanhq import DhanHTTP


class Security:

    """"Constants for HTTP Responses"""
    OTP_SENT = 'OTP sent'

    """CSV URL for Security ID List"""
    COMPACT_CSV_URL = 'https://images.dhan.co/api-data/api-scrip-master.csv'
    DETAILED_CSV_URL = 'https://images.dhan.co/api-data/api-scrip-master-detailed.csv'

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def _save_as_temp_html_file_and_open_in_browser(self, form_html):
        temp_web_form_html = "temp_form.html"
        with open(temp_web_form_html, "w") as f:
            f.write(form_html)
        web_open(Path.cwd().joinpath(temp_web_form_html).as_uri())

    def generate_tpin(self):
        """
        Generate T-Pin on registered mobile number.

        Returns:
            dict: The response containing the status of T-Pin generation.
        """
        endpoint = '/edis/tpin'
        response = self.dhan_http.get(endpoint)
        response['data'] = ''
        # ToDo: This is inconsistent. If success then data should be set and not remarks field
        if response['status'] == DhanHTTP.HttpResponseStatus.SUCCESS.value:
            response['remarks'] = self.OTP_SENT
        else:
            # ToDo: Why this redundant code here?
            response['remarks'] = 'status code : ' + response['remarks']['error_code']
        return response

    def open_browser_for_tpin(self, isin, qty, exchange, segment='EQ', bulk=False):
        """
        Opens the default web browser to enter T-Pin.

        Args:
            isin (str): The ISIN of the security.
            qty (int): The quantity of the security.
            exchange (str): The exchange where the security is listed.
            segment (str): The segment of the exchange (default is 'EQ').
            bulk (bool): Flag for bulk operations (default is False).

        Returns:
            dict: The response containing the status of the operation.
        """
        endpoint = '/edis/form'
        payload = {
            "isin": isin,
            "qty": qty,
            "exchange": exchange,
            "segment": segment,
            "bulk": bulk
        }
        response = self.dhan_http.post(endpoint, payload)
        if response['status'] == DhanHTTP.HttpResponseStatus.FAILURE.value:
            return response

        data = json_loads(response['data'])
        form_html = data['edisFormHtml']  # data['edisFormHtml']
        form_html = form_html.replace('\\', '')
        # print(form_html)
        self._save_as_temp_html_file_and_open_in_browser(form_html)
        return response

    def edis_inquiry(self, isin):
        """
        Inquire about the eDIS status of the provided ISIN.

        Args:
            isin (str): The ISIN to inquire about.

        Returns:
            dict: The response containing inquiry results.
        """
        endpoint = f'/edis/inquire/{isin}'
        return self.dhan_http.get(endpoint)

    @staticmethod
    def fetch_security_list(mode='compact', filename='security_id_list.csv'):
        """
        Fetch CSV file from dhan based on the specified mode and save it to the current directory.

        Args:
            mode (str): The mode to fetch the CSV ('compact' or 'detailed').
            filename (str): The name of the file to save the CSV as (default is 'data.csv').

        Returns:
            pd.DataFrame: The DataFrame containing the CSV data.
        """
        import pandas as pd
        try:
            if mode == 'compact':
                csv_url = Security.COMPACT_CSV_URL
            elif mode == 'detailed':
                csv_url = Security.DETAILED_CSV_URL
            else:
                raise ValueError("Invalid mode. Choose 'compact' or 'detailed'.")

            response = requests.get(csv_url)
            response.raise_for_status()

            with open(filename, 'wb') as f:
                f.write(response.content)
            df = pd.read_csv(filename)
            return df
        except Exception as e:
            logging.error('Exception in dhanhq>>fetch_security_list: %s', e)
            return None



from json import dumps as json_dumps
from unittest.mock import patch

import pytest

from dhanhq import DhanContext
from dhanhq.dhan_http import DhanHTTP
from dhanhq.dhanhq import dhanhq


class TestDhanhq_Orders:
    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_order_list_success(self, mock_read_request, dhanhq_obj):
        dhanhq_obj.get_order_list()
        mock_read_request.assert_called_once_with('/orders')

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_order_by_id(self, mock_read_request, dhanhq_obj):
        order_id = "12345"
        dhanhq_obj.get_order_by_id(order_id)
        mock_read_request.assert_called_once_with(f'/orders/{order_id}')

    @patch("dhanhq.dhan_http.DhanHTTP.get")
    def test_get_order_by_correlation_id(self, mock_read_request, dhanhq_obj):
        correlation_id = "12345"
        dhanhq_obj.get_order_by_correlationID(correlation_id)
        mock_read_request.assert_called_once_with(f'/orders/external/{correlation_id}')

    @patch("dhanhq.dhan_http.DhanHTTP.delete")
    def test_cancel_order_success(self, mock_delete_request, dhanhq_obj):
        order_id = "12345"
        dhanhq_obj.cancel_order(order_id)
        mock_delete_request.assert_called_once_with(f'/orders/{order_id}')

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_place_order_success(self,mock_create_request,dhanhq_obj):
        endpoint = '/orders'
        security_id = 1
        exchange_segment = "exchange_segment"
        transaction_type = "transaction_type"
        quantity =100
        order_type = "order_type"
        product_type = "product_type"
        price = 123
        dhanhq_obj.place_order(security_id, exchange_segment, transaction_type, quantity,
                    order_type, product_type, price, trigger_price=0, disclosed_quantity=0,
                    after_market_order=False, validity='DAY', amo_time='OPEN',
                    bo_profit_value=None, bo_stop_loss_Value=None, tag=None)
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.post")
    def test_place_slice_order_success(self,mock_create_request,dhanhq_obj):
        endpoint = '/orders/slicing'
        security_id = 1
        exchange_segment = "exchange_segment"
        transaction_type = "transaction_type"
        quantity =100
        order_type = "order_type"
        product_type = "product_type"
        price = 123
        dhanhq_obj.place_slice_order(security_id, exchange_segment, transaction_type, quantity,
                    order_type, product_type, price, trigger_price=0, disclosed_quantity=0,
                    after_market_order=False, validity='DAY', amo_time='OPEN',
                    bo_profit_value=None, bo_stop_loss_Value=None, tag=None)
        mock_create_request.assert_called_once()
        assert mock_create_request.call_args[0][0] == endpoint

    @patch("dhanhq.dhan_http.DhanHTTP.put")
    def test_modify_order_success(self,mock_update_request,dhanhq_obj):
        order_id = 123
        endpoint = f'/orders/{order_id}'
        quantity = 100
        price = 99
        trigger_price = 100
        dhanhq_obj.modify_order(order_id, "order_type", "leg_name", quantity, price,
                                trigger_price, trigger_price, "validity")
        mock_update_request.assert_called_once()
        assert mock_update_request.call_args[0][0] == endpoint




import json
import os
from pprint import pprint

import pytest
from dhanhq import dhanhq

# from tests.conftest import api_access_token_fixture

class TestE2E_Dhanhq_Orders:
    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get-current-orders-list.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_order_list(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_order_list()
        # pprint(actual_response)
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get-order-by-id.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_order_by_id(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_order_by_id("string")
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get-order-by-correlation-id.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_order_by_correlation_id(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_order_by_id("string")
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/modify_pending_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_modify_order(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.modify_order("string",dhanhq.LIMIT, "ENTRY_LEG",1,11,12.0,1,dhanhq.DAY)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/cancel_given_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_cancel_order(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.cancel_order("string")
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/place_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_place_order(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.place_order("security_id", dhanhq.NSE, dhanhq.BUY,
                                                     100, dhanhq.LIMIT, dhanhq.CNC, 99.99, 0)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/place_slice_order.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_place_slice_order(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.place_slice_order("security_id", dhanhq.NSE, dhanhq.BUY,
                                                     100, dhanhq.LIMIT, dhanhq.CNC, 99.99, 0)
        # assert actual_response['status'] == expected_dict['status']
        # assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']





class MarketFeed:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def ticker_data(self, securities):
        """
        Retrieve the latest market price for specified instruments.

        Args:
            securities (dict): A dictionary where keys are exchange segments and values are lists of security IDs.
                securities = {
                    "NSE_EQ": [11536],
                    "NSE_FNO": [49081, 49082]
                }

        Returns:
            dict: The response containing last traded price (LTP) data.
        """
        endpoint = '/marketfeed/ltp'
        payload = {exchange_segment: security_id for exchange_segment, security_id in securities.items()}
        return self.dhan_http.post(endpoint, payload)

    def ohlc_data(self, securities):
        """
        Retrieve the Open, High, Low and Close price along with LTP for specified instruments.

        Args:
            securities (dict): A dictionary where keys are exchange segments and values are lists of security IDs.
                securities = {
                    "NSE_EQ": [11536],
                    "NSE_FNO": [49081, 49082]
                }

        Returns:
            dict: The response containing Open, High, Low and Close along with LTP data.
        """
        endpoint = '/marketfeed/ohlc'
        payload = {exchange_segment: security_id for exchange_segment, security_id in securities.items()}
        return self.dhan_http.post(endpoint, payload)

    def quote_data(self, securities):
        """
        Retrieve full details including market depth, OHLC data, OI and volume along with LTP for specified instruments.

        Args:
            securities (dict): A dictionary where keys are exchange segments and values are lists of security IDs.
                securities = {
                    "NSE_EQ": [11536],
                    "NSE_FNO": [49081, 49082]
                }

        Returns:
            dict: The response containing full packet including market depth, last trade,
                    circuit limit, OHLC, OI and volume data.
        """
        endpoint = '/marketfeed/quote'
        payload = {exchange_segment: security_id for exchange_segment, security_id in securities.items()}
        return self.dhan_http.post(endpoint, payload)




import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '2.1.0'
PACKAGE_NAME = 'dhanhq'
AUTHOR = 'Dhan'
AUTHOR_EMAIL = 'dhan-oss@dhan.co'
URL = 'https://dhanhq.co/'

LICENSE = 'MIT LICENSE'
DESCRIPTION = 'The official Python client for communicating with the DhanHQ API'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
    "pandas>=1.4.3",
    "requests>=2.28.1",
    "websockets>=12.0.1",
    "pyOpenSSL>=20.0.1",
    ]

TEST_SUITE="tests"
TEST_REQUIRES = [
    "python-dotenv>=1.0.1",
    "flake8>=7.1.1",
    "pylint>=3.3.3",
    "pytest>=8.3.4",
    "pytest-cov>=6.0.0",
    "responses>=0.25.3",
    ]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      package_dir={'': 'src'}, # Tells setuptools where your code is
      packages=find_packages(where='src'), # Automatically find packages
      install_requires=INSTALL_REQUIRES,
      extras_require={
          'dev': TEST_REQUIRES
      },
      )



"""
    A class to interact with the DhanHQ APIs using HTTP protocol.

    This library provides methods to manage orders, retrieve market data,
    and perform various trading operations through the DhanHQ API.

    :copyright: (c) 2025 by Dhan.
    :license: see LICENSE for details.
"""

import logging
from enum import Enum
from json import dumps as json_dumps, loads as json_loads

import requests


class DhanHTTP:
    """Manages API keys, connection context, and HTTP requests"""

    class HttpResponseStatus(Enum):
        """Constants for HTTP Status Codes"""
        SUCCESS = 'success'
        FAILURE = 'failure'

    class HttpMethods(Enum):
        """Constants for HTTP Requests"""
        GET = 'GET'
        POST = 'POST'
        PUT = 'PUT'
        DELETE = 'DELETE'

    HTTP_DEFAULT_TIME_OUT = 60
    API_BASE_URL = 'https://api.dhan.co/v2'

    def __init__(self, client_id, access_token, disable_ssl=False, pool=None):
        self.client_id = client_id
        self.access_token = access_token
        self.base_url = DhanHTTP.API_BASE_URL
        self.timeout = DhanHTTP.HTTP_DEFAULT_TIME_OUT
        self.header = {
            'access-token': self.access_token,
            'client-id': self.client_id,
            'Content-type': 'application/json',
            'Accept': 'application/json'
        }
        self.disable_ssl = disable_ssl
        self.session = requests.Session()
        if pool:
            reqadapter = requests.adapters.HTTPAdapter(**pool)
            self.session.mount("https://", reqadapter)

    def _send_request(self, method, endpoint, payload=None):
        url = self.base_url + endpoint
        if payload:
            payload["dhanClientId"] = self.client_id
            payload = json_dumps(payload)
        try:
            response = getattr(self.session, method.value.lower())(url,
                                                                   data=payload,
                                                                   headers=self.header,
                                                                   timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in DhanHQConnection.%s: %s', method.value.upper(), e)
            return {
                'status': DhanHTTP.HttpResponseStatus.FAILURE.value,
                'remarks': str(e),
                'data': '',
            }

    def _parse_response(self, response):
        """
        Parse the API's string response to return a JSON as dict.

        Args:
            response (requests.Response): The response object from the API.

        Returns:
            dict: Parsed response containing status, remarks, and data.
        """
        try:
            status = DhanHTTP.HttpResponseStatus.FAILURE.value
            remarks = ''
            data = ''
            json_response = json_loads(response.content)
            if (response.status_code >= 200) and (response.status_code <= 299):
                status = DhanHTTP.HttpResponseStatus.SUCCESS.value
                data = json_response
            else:
                remarks = {
                    'error_code': (json_response.get('errorCode')),
                    'error_type': (json_response.get('errorType')),
                    'error_message': (json_response.get('errorMessage'))
                }
        except Exception as e:
            logging.warning('Exception found in dhanhq>>find_error_code: %s', e)
            status = DhanHTTP.HttpResponseStatus.FAILURE.value
            remarks = str(e)
        return {
            'status': status,
            'remarks': remarks,
            'data': data,
        }

    def get(self, endpoint):
        """
        Do HTTP-GET request to Dhan Endpoint.

        Args:
            endpoint (str): The endpoint ignoring the base URL.

        Returns:
        dict: The response in dict format.
        """
        return self._send_request(DhanHTTP.HttpMethods.GET, endpoint)

    def post(self, endpoint, payload):
        """
        Do HTTP-POST request to Dhan Endpoint.

        Args:
            endpoint (str): The endpoint ignoring the base URL.
            payload (dict): The payload dict contains the data that needs to be sent to the server.

        Returns:
        dict: The response in dict format.
        """
        return self._send_request(DhanHTTP.HttpMethods.POST, endpoint, payload)

    def put(self, endpoint, payload):
        """
        Do HTTP-PUT request to Dhan Endpoint.

        Args:
            endpoint (str): The endpoint ignoring the base URL.
            payload (dict): The payload dict contains the data that needs to be sent to the server.

        Returns:
        dict: The response in dict format.
        """
        return self._send_request(DhanHTTP.HttpMethods.PUT, endpoint, payload)

    def delete(self, endpoint):
        """
        Do HTTP-DELETE request to Dhan Endpoint.

        Args:
            endpoint (str): The endpoint ignoring the base URL.

        Returns:
        dict: The response in dict format.
        """
        return self._send_request(DhanHTTP.HttpMethods.DELETE, endpoint)




class Order:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def get_order_list(self):
        """
        Retrieve a list of all orders requested in a day with their last updated status.

        Returns:
            dict: The response containing order list status and data.
        """
        return self.dhan_http.get('/orders')

    def get_order_by_id(self, order_id):
        """
        Retrieve the details and status of an order from the orderbook placed during the day.

        Args:
            order_id (str): The ID of the order to retrieve.

        Returns:
            dict: The response containing order details and status.
        """
        return self.dhan_http.get(f'/orders/{order_id}')

    def get_order_by_correlationID(self, correlation_id):
        """
        Retrieve the order status using a field called correlation_id.

        Args:
            correlation_id (str): The correlation_id provided during order placement.

        Returns:
            dict: The response containing order status.
        """
        return self.dhan_http.get(f'/orders/external/{correlation_id}')

    def modify_order(self, order_id, order_type, leg_name, quantity,
                     price, trigger_price, disclosed_quantity, validity):
        """
        Modify a pending order in the orderbook.

        Args:
            order_id (str): The ID of the order to modify.
            order_type (str): The type of order (e.g., LIMIT, MARKET).
            leg_name (str): The name of the leg to modify.
            quantity (int): The new quantity for the order.
            price (float): The new price for the order.
            trigger_price (float): The trigger price for the order.
            disclosed_quantity (int): The disclosed quantity for the order.
            validity (str): The validity of the order.

        Returns:
            dict: The response containing the status of the modification.
        """
        payload = {
            "orderId": str(order_id),
            "orderType": order_type,
            "legName": leg_name,
            "quantity": quantity,
            "price": price,
            "disclosedQuantity": disclosed_quantity,
            "triggerPrice": trigger_price,
            "validity": validity
        }
        return self.dhan_http.put(f'/orders/{order_id}', payload)

    def cancel_order(self, order_id):
        """
        Cancel a pending order in the orderbook using the order ID.

        Args:
            order_id (str): The ID of the order to cancel.

        Returns:
            dict: The response containing the status of the cancellation.
        """
        return self.dhan_http.delete(f'/orders/{order_id}')

    def place_order(self, security_id, exchange_segment, transaction_type, quantity,
                    order_type, product_type, price, trigger_price=0, disclosed_quantity=0,
                    after_market_order=False, validity='DAY', amo_time='OPEN',
                    bo_profit_value=None, bo_stop_loss_Value=None, tag=None, should_slice=False):
        """
        Place a new order in the Dhan account.

        Args:
            security_id (str): The ID of the security to trade.
            exchange_segment (str): The exchange segment (e.g., NSE, BSE).
            transaction_type (str): The type of transaction (BUY/SELL).
            quantity (int): The quantity of the order.
            order_type (str): The type of order (LIMIT, MARKET, etc.).
            product_type (str): The product type (CNC, INTRA, etc.).
            price (float): The price of the order.
            trigger_price (float): The trigger price for the order.
            disclosed_quantity (int): The disclosed quantity for the order.
            after_market_order (bool): Flag for after market order.
            validity (str): The validity of the order (DAY, IOC, etc.).
            amo_time (str): The time for AMO orders.
            bo_profit_value (float): The profit value for BO orders.
            bo_stop_loss_Value (float): The stop loss value for BO orders.
            tag (str): Optional correlation ID for tracking.

        Returns:
            dict: The response containing the status of the order placement.
        """

        if after_market_order and (amo_time not in ['OPEN', 'OPEN_30', 'OPEN_60']):
            raise Exception("amo_time value must be one of ['OPEN','OPEN_30','OPEN_60']")

        payload = {
            "transactionType": transaction_type.upper(),
            "exchangeSegment": exchange_segment.upper(),
            "productType": product_type.upper(),
            "orderType": order_type.upper(),
            "validity": validity.upper(),
            "securityId": security_id,
            "quantity": int(quantity),
            "disclosedQuantity": int(disclosed_quantity),
            "price": float(price),
            "afterMarketOrder": after_market_order,
            "boProfitValue": bo_profit_value,
            "boStopLossValue": bo_stop_loss_Value,
            "triggerPrice": float(trigger_price)
        }

        if tag is not None and tag != '':
            payload["correlationId"] = tag

        endpoint = '/orders'
        if should_slice:
            endpoint += '/slicing'
        return self.dhan_http.post(endpoint, payload)

    def place_slice_order(self, security_id, exchange_segment, transaction_type, quantity,
                          order_type, product_type, price, trigger_price=0, disclosed_quantity=0,
                          after_market_order=False, validity='DAY', amo_time='OPEN',
                          bo_profit_value=None, bo_stop_loss_Value=None, tag=None):
        """
        Place a new slice order in the Dhan account.

        Args:
            security_id (str): The ID of the security to trade.
            exchange_segment (str): The exchange segment (e.g., NSE, BSE).
            transaction_type (str): The type of transaction (BUY/SELL).
            quantity (int): The quantity of the order.
            order_type (str): The type of order (LIMIT, MARKET, etc.).
            product_type (str): The product type (CNC, MIS, etc.).
            price (float): The price of the order.
            trigger_price (float): The trigger price for the order.
            disclosed_quantity (int): The disclosed quantity for the order.
            after_market_order (bool): Flag for after market order.
            validity (str): The validity of the order (DAY, IOC, etc.).
            amo_time (str): The time for AMO orders.
            bo_profit_value (float): The profit value for BO orders.
            bo_stop_loss_Value (float): The stop loss value for BO orders.
            tag (str): Optional correlation ID for tracking.

        Returns:
            dict: The response containing the status of the slice order placement.
        """

        return self.place_order(security_id, exchange_segment, transaction_type, quantity,
                                order_type, product_type, price, trigger_price, disclosed_quantity,
                                after_market_order, validity, amo_time,
                                bo_profit_value, bo_stop_loss_Value, tag, True)



import json
import os
from datetime import date
from pprint import pprint

import pytest


class TestE2E_Dhanhq_StatementsAndReports:
    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_all_trades.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_trade_book_without_orderid(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_trade_book()
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_trade_book_by_orderid.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_trade_book_by_orderid(self, expected_dict, dhanhq_fixture):
        actual_response = dhanhq_fixture.get_trade_book("order_id")
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']

    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_trade_history.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_trade_history(self, expected_dict, dhanhq_fixture):
        str_today = date.today().strftime('%Y-%m-%d')
        actual_response = dhanhq_fixture.get_trade_history(str_today,str_today,page_number=0)
        # pprint(actual_response)
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']


    json_file_path = os.path.join(os.path.dirname(__file__), '../data/get_ledger_report.json')
    @pytest.mark.parametrize("expected_dict", [json.load(open(json_file_path))])
    def test_get_ledger_report(self, expected_dict, dhanhq_fixture):
        str_today = date.today().strftime('%Y-%m-%d')
        actual_response = dhanhq_fixture.ledger_report(str_today, str_today)
        # pprint(actual_response)
        assert actual_response['status'] == expected_dict['status']
        assert actual_response['remarks'] == expected_dict['remarks']
        # assert actual_response['data'] == expected_dict['data']





class OptionChain:

    def __init__(self, dhan_context):
        self.dhan_http = dhan_context.get_dhan_http()

    def option_chain(self, under_security_id, under_exchange_segment, expiry):
        """
        Retrieve the real-time Option Chain for a specified underlying instrument.

        Args:
            under_security_id (int): The security ID of the underlying instrument.
            under_exchange_segment (str): The exchange segment of the underlying instrument (e.g., NSE, BSE).
            expiry (str): The expiry date of the options.

        Returns:
            dict: The response containing Open Interest (OI), Greeks, Volume, Last Traded Price,
                    Best Bid/Ask, and Implied Volatility (IV) across all strikes for the specified underlying.
        """
        endpoint = '/optionchain'
        payload = {
            "UnderlyingScrip": under_security_id,
            "UnderlyingSeg": under_exchange_segment,
            "Expiry": expiry
        }
        return self.dhan_http.post(endpoint, payload)

    def expiry_list(self, under_security_id, under_exchange_segment):
        """
        Retrieve the dates of all expiries for a specified underlying instrument.

        Args:
            under_security_id (int): The security ID of the underlying instrument.
            under_exchange_segment (str): The exchange segment of the underlying instrument (e.g., NSE, BSE).

        Returns:
            dict: The response containing list of dates for which option expiries
                    are present for the specified underlying instrument.
        """
        endpoint = '/optionchain/expirylist'
        payload = {
            "UnderlyingScrip": under_security_id,
            "UnderlyingSeg": under_exchange_segment
        }
        return self.dhan_http.post(endpoint, payload)




"""
    The marketfeed class is designed to facilitate asynchronous communication with the DhanHQ API via WebSocket.
    It enables users to subscribe to market data for a list of instruments and receive real-time updates.

    :copyright: (c) 2025 by Dhan.
    :license: see LICENSE for details.
"""

import websockets
import asyncio
import struct
from datetime import datetime
from collections import defaultdict
import json


class MarketFeed:
    # Constants
    """WebSocket URL for DhanHQ Live Market Feed"""
    market_feed_wss = 'wss://api-feed.dhan.co'

    """Constants for Exchange Segment"""
    IDX = 0
    NSE = 1
    NSE_FNO = 2
    NSE_CURR = 3
    BSE = 4
    MCX = 5
    BSE_CURR = 7
    BSE_FNO = 8

    """Constants for Request Code"""
    Ticker = 15
    Quote = 17
    Depth = 19
    Full = 21

    def __init__(self, dhan_context, instruments, version='v1'):
        """Initializes the MarketFeed instance with user credentials, instruments to subscribe, and callback functions."""

        self.client_id = dhan_context.get_client_id()
        self.access_token = dhan_context.get_access_token()
        self.instruments = instruments
        self.data = ""
        self._is_first_connect = True
        self.ws = None
        self.on_ticks = None
        self.loop = asyncio.get_event_loop()
        self.version = version

    def run_forever(self):
        """Starts the WebSocket connection and runs the event loop."""
        self.loop.run_until_complete(self.connect())

    def get_data(self):
        """Fetch instruments data while the event loop is open."""
        return self.loop.run_until_complete(self.get_instrument_data())

    def close_connection(self):
        """Close WebSocket connection with this."""
        return self.loop.run_until_complete(self.disconnect())

    async def connect(self):
        """Initiates the connection to the Websockets."""
        if not self.ws or self.ws.state == websockets.protocol.State.CLOSED:
            if self.version == 'v1':
                self.ws = await websockets.connect(MarketFeed.market_feed_wss)
                await self.authorize()
            elif self.version == 'v2':
                url = f"{MarketFeed.market_feed_wss}?version=2&token={self.access_token}&clientId={self.client_id}&authType=2"
                self.ws = await websockets.connect(url)
            else:
                raise ValueError(f"Unsupported version: {self.version}")
            await self.subscribe_instruments()
        else:
            try:
                await self.ws.ping()
            except websockets.ConnectionClosed:
                self.ws = None
                await self.connect()

    async def get_instrument_data(self):
        """Fetches data and initiates process for conversion."""
        response = await self.ws.recv()
        self.data = self.process_data(response)
        return self.data

    async def disconnect(self):
        """Closes the WebSocket connection and sends a disconnect message."""
        if self.ws:
            if self.version == 'v2':
                disconnect_message = {
                    "RequestCode": 12
                }
                await self.ws.send(json.dumps(disconnect_message))
                header_message = self.create_header(feed_request_code=12, message_length=83, client_id=self.client_id)
                await self.ws.send(header_message)
        print("Connection closed!")

    async def authorize(self):
        """Establishes the WebSocket connection and authorizes the user"""
        if self.version != 'v1':
            print("Authorization not needed for this version.")
            self.is_authorized = True
            return
        try:
            print("Authorizing...")

            # Authorization packet creation
            api_access_token = self.access_token.encode('utf-8')
            api_access_token = self.pad_with_zeros(api_access_token, 500)
            authentication_type = "2P".encode('utf-8')
            payload = api_access_token + authentication_type

            feed_request_code = 11
            message_length = 83 + len(api_access_token) + len(authentication_type)
            client_id = self.client_id.encode('utf-8')
            client_id = self.pad_with_zeros(client_id, 30)
            dhan_auth = b"\0" * 50
            header = struct.pack('<bH30s50s', feed_request_code, message_length, client_id, dhan_auth)

            authorization_packet = header + payload

            # Send authorization packet
            await self.ws.send(authorization_packet)
            self.is_authorized = True
            print("Authorization successful!")
        except Exception as e:
            print(f"Authorization failed: {e}")
            self.is_authorized = False


    """Creating Instruments List to be subscribed"""
    def validate_and_process_tuples(self, tuples_list, batch_size=100):
        """Create a list of all instruments to be added and add in batches of 100"""
        # Check if all tuples have the same size (either all size 2 or all size 3)
        tuple_lengths = set(len(tup) for tup in tuples_list)
        if len(tuple_lengths) > 1:
            raise ValueError("All tuples must be of the same size, either all 2 or all 3.")

        # Assign default type 15 to tuples of size 2
        processed_tuples = []
        for tup in tuples_list:
            if len(tup) == 2:
                processed_tuples.append((tup[0], tup[1], 15))
            else:
                processed_tuples.append(tup)

        # Eliminate duplicate tuples
        processed_tuples = list(set(processed_tuples))

        # Separate tuples by type and prepare for batching
        batches = defaultdict(list)
        for tup in processed_tuples:
            exchange, instrument_id, type_ = tup
            if self.version == 'v1' and type_ not in [15, 17, 19]:
                raise ValueError(f"Invalid request mode for v1. Only Ticker, Quote, and Depth packet are allowed.")
            if self.version == 'v2' and type_ not in [15, 17, 21]:
                raise ValueError(f"Invalid request mode for v2. Only Ticker, Quote, and Full packet are allowed.")
            batches[type_].append((exchange, instrument_id))

        # Create the final dictionary with batches of the specified size
        final_dict = {}
        for type_ in [15, 17, 19, 21]:
            type_batches = [batches[type_][i:i+batch_size] for i in range(0, len(batches[type_]), batch_size)]
            final_dict[str(type_)] = type_batches

        return final_dict

    async def subscribe_instruments(self):
        """Subscribe Instruments on the Open Websocket"""
        if self.version == 'v1':
            if not self.is_authorized:
                print("Not authorized. Please authorize first.")
                return

            # Subscription packet creation
            group_size = 100

            new_instrument_list = self.validate_and_process_tuples(self.instruments, group_size)

            if new_instrument_list != {}:
                for instrument_type in new_instrument_list:
                    if new_instrument_list[instrument_type] != []:
                        for instrument_group in new_instrument_list[instrument_type]:
                            self.subscription_code = int(instrument_type)
                            subscription_packet = self.create_subscription_packet(instrument_group, self.subscription_code)
                            await self.ws.send(subscription_packet)
        elif self.version == 'v2':
            new_instrument_list = self.validate_and_process_tuples(self.instruments)
            for instrument_type, instrument_groups in new_instrument_list.items():
                for instrument_group in instrument_groups:
                    # Split the instrument group into batches of 100
                    for i in range(0, len(instrument_group), 100):
                        batch = instrument_group[i:i+100]
                        subscription_message = {
                            "RequestCode": int(instrument_type),
                            "InstrumentCount": len(batch),
                            "InstrumentList": [
                                {
                                    "ExchangeSegment": self.get_exchange_segment(ex),
                                    "SecurityId": token
                                } for ex, token in batch
                            ]
                        }
                        await self.ws.send(json.dumps(subscription_message))

    def get_exchange_segment(self, exchange_code):
        """Convert numeric exchange code to string representation"""
        exchange_map = {
            0: "IDX_I",
            1: "NSE_EQ",
            2: "NSE_FNO",
            3: "NSE_CURRENCY",
            4: "BSE_EQ",
            5: "MCX_COMM",
            7: "BSE_CURRENCY",
            8: "BSE_FNO"
        }
        return exchange_map.get(exchange_code, str(exchange_code))

    def process_data(self, data):
        """Read binary data and initiate processing in received format"""
        first_byte = struct.unpack('<B', data[0:1])[0]
        self.on_close = False
        if first_byte == 2:
            return self.process_ticker(data)
        elif first_byte == 3:
            return self.process_market_depth(data)
        elif first_byte == 4:
            return self.process_quote(data)
        elif first_byte == 5:
            return self.process_oi(data)
        elif first_byte == 6:
            return self.process_prev_close(data)
        elif first_byte == 7:
            return self.process_status(data)
        elif first_byte == 8:
            return self.process_full(data)
        elif first_byte == 50:
            return self.server_disconnection(data)

    def process_ticker(self, data):
        """Parse and process Ticker Data"""
        unpack_ticker = [struct.unpack('<BHBIfI', data[0:16])]
        # Assign unpacked data to fields
        ticker_data = {
            "type" : 'Ticker Data',
            "exchange_segment" : unpack_ticker[0][2],
            "security_id" : unpack_ticker[0][3],
            "LTP" : "{:.2f}".format(unpack_ticker[0][4]),
            "LTT" : self.utc_time(unpack_ticker[0][5])
        }
        return ticker_data

    def process_prev_close(self, data):
        """Parse and process Previous Day Data"""
        unpack_pclose = [struct.unpack('<BHBIfI', data[0:16])]
        # Assign unpacked data to fields
        prev_close = {
            "type" : 'Previous Close',
            "exchange_segment" : unpack_pclose[0][2],
            "security_id" : unpack_pclose[0][3],
            "prev_close" : "{:.2f}".format(unpack_pclose[0][4]),
            "prev_OI" : unpack_pclose[0][5]
        }
        return prev_close

    def process_market_depth(self, data):
        """Parse and process Market Depth Data"""
        market_data = [struct.unpack('<BHBIf100s', data[0:112])]
        market_data_list = list(market_data[0][0:5])

        exchange_segment = market_data_list[2]
        sec_id = market_data_list[3]
        ltp = market_data_list[4]

        # Process Market Depth
        market_depth_binary = market_data[0][5]
        packet_format = '<IIHHff'
        packet_size = struct.calcsize(packet_format)
        depth = []
        for i in range(5):
            start_idx = i * packet_size
            end_idx = start_idx + packet_size
            current_packet = struct.unpack(packet_format, market_depth_binary[start_idx:end_idx])

            depth.append({
                "bid_quantity": current_packet[0],
                "ask_quantity": current_packet[1],
                "bid_orders": current_packet[2],
                "ask_orders": current_packet[3],
                "bid_price": "{:.2f}".format(current_packet[4]),
                "ask_price": "{:.2f}".format(current_packet[5])
            })

        market_depth = {
            "type" : 'Market Depth',
            "exchange_segment" : exchange_segment,
            "security_id" : sec_id,
            "LTP" : ltp,
            "depth" : depth
                   }
        return market_depth

    def process_quote(self, data):
        """Parse and process Quote Data"""
        unpack_quote = [struct.unpack('<BHBIfHIfIIIffff', data[0:50])]
        quote_data = {
            "type" : 'Quote Data',
            "exchange_segment" : unpack_quote[0][2],
            "security_id": unpack_quote[0][3],
            "LTP": "{:.2f}".format(unpack_quote[0][4]),
            "LTQ" : unpack_quote[0][5],
            "LTT" : self.utc_time(unpack_quote[0][6]),
            "avg_price" : "{:.2f}".format(unpack_quote[0][7]),
            "volume" : unpack_quote[0][8],
            "total_sell_quantity" : unpack_quote[0][9],
            "total_buy_quantity" : unpack_quote[0][10],
            "open" : "{:.2f}".format(unpack_quote[0][11]),
            "close": "{:.2f}".format(unpack_quote[0][12]),
            "high": "{:.2f}".format(unpack_quote[0][13]),
            "low": "{:.2f}".format(unpack_quote[0][14])
        }
        return quote_data

    def process_oi(self, data):
        """Parse and process OI Data"""
        unpack_oi = [struct.unpack('<BHBII', data[0:12])]
        oi_data = {
            "type" : 'OI Data',
            "exchange_segment" : unpack_oi[0][2],
            "security_id": unpack_oi[0][3],
            "OI": unpack_oi[0][4]
        }
        return oi_data

    def process_status(self, data):
        """Parse and process market status"""
        unpack_status = [struct.unpack('<BHBI', data[0:8])]
        market_status = "Markets Open"
        return market_status

    def process_full(self, data):
        """Parse and process Full Packet Data"""
        unpack_full = [struct.unpack('<BHBIfHIfIIIIIIffff100s', data[0:162])]

        # Processing Market Depth in Full Packet
        market_depth_binary = unpack_full[0][18]
        packet_format = '<IIHHff'
        packet_size = struct.calcsize(packet_format)

        depth = []
        for i in range(5):
            start_idx = i * packet_size
            end_idx = start_idx + packet_size
            current_packet = struct.unpack(packet_format, market_depth_binary[start_idx:end_idx])

            depth.append({
                "bid_quantity": current_packet[0],
                "ask_quantity": current_packet[1],
                "bid_orders": current_packet[2],
                "ask_orders": current_packet[3],
                "bid_price": "{:.2f}".format(current_packet[4]),
                "ask_price": "{:.2f}".format(current_packet[5])
            })

        full_packet = {
            "type" : 'Full Data',
            "exchange_segment" : unpack_full[0][2],
            "security_id": unpack_full[0][3],
            "LTP": "{:.2f}".format(unpack_full[0][4]),
            "LTQ" : unpack_full[0][5],
            "LTT" : self.utc_time(unpack_full[0][6]),
            "avg_price" : "{:.2f}".format(unpack_full[0][7]),
            "volume" : unpack_full[0][8],
            "total_sell_quantity" : unpack_full[0][9],
            "total_buy_quantity" : unpack_full[0][10],
            "OI" : unpack_full[0][11],
            "oi_day_high" : unpack_full[0][12],
            "oi_day_low" : unpack_full[0][13],
            "open" : "{:.2f}".format(unpack_full[0][14]),
            "close": "{:.2f}".format(unpack_full[0][15]),
            "high": "{:.2f}".format(unpack_full[0][16]),
            "low": "{:.2f}".format(unpack_full[0][17]),
            "depth": depth
        }
        return full_packet


    def server_disconnection(self, data):
        """Parse and process server disconnection error"""
        disconnection_packet = [struct.unpack('<BHBIH', data[0:10])]
        self.on_close = False
        if disconnection_packet[0][4] == 805:
            print ("Disconnected: No. of active websocket connections exceeded")
            self.on_close = True
        elif disconnection_packet[0][4] == 806:
            print ("Disconnected: Subscribe to Data APIs to continue")
            self.on_close = True
        elif disconnection_packet[0][4] == 807:
            print ("Disconnected: Access Token is expired")
            self.on_close = True
        elif disconnection_packet[0][4] == 808:
            print ("Disconnected: Invalid Client ID")
            self.on_close = True
        elif disconnection_packet[0][4] == 809:
            print ("Disconnected: Authentication Failed - check ")
            self.on_close = True

    async def on_connection_opened(self, websocket):
        "Callback function executed when the WebSocket connection is opened."
        await websocket.send(self.create_subscription_packet(self.instruments, subscribe=True))

        while True:
            response = await websocket.recv()
            await self.on_message_received(response)

    def pad_with_zeros(self, data, length):
        """Pads a binary data string with zeros to a specified length for server to read."""
        data = data.ljust(length, b'\0')
        return data

    def create_header(self, feed_request_code, message_length, client_id):
        """Creates header packet for the subscription packet."""
        dhan_auth = b"\0" * 50
        header = struct.pack('<bH30s50s', feed_request_code, message_length, client_id.encode('utf-8'), dhan_auth)
        return header

    def utc_time(self, epoch_time):
        """Converts EPOCH time to UTC time."""
        return datetime.utcfromtimestamp(epoch_time).strftime('%H:%M:%S')

    def create_subscription_packet(self, instruments, feed_request_code):
        """Creates the subscription packet with specified instruments and subscription code"""
        num_instruments = len(instruments)

        header = self.create_header(feed_request_code = feed_request_code,
                                    message_length=83 + 4 + num_instruments * 21,
                                    client_id=self.client_id)
        num_instruments_bytes = struct.pack('<I', num_instruments)
        instrument_info = b""
        for exchange_segment, security_id in instruments:
            instrument_info += struct.pack('<B20s', exchange_segment, security_id.encode('utf-8'))

        instruments = [(0, "")]
        for i in range(100 - num_instruments):
            instrument_info += struct.pack('<B20s', instruments[0][0], instruments[0][1].encode('utf-8'))

        subscription_packet = header + num_instruments_bytes + instrument_info
        return subscription_packet

    def subscribe_symbols(self, symbols):
        """Function to subscribe to additional symbols when connection is already established."""
        # Update the instruments list
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.update(symbols)
        self.instruments = list(unique_symbols_set)

        # If the WebSocket is open, send the subscription packet for the new symbols
        if self.ws and not self.ws.closed:
            # Prepare the instruments list for subscription
            group_size = 100
            new_instrument_list = self.validate_and_process_tuples(symbols, group_size)

            if self.version == 'v1':
                for instrument_type in new_instrument_list:
                    if new_instrument_list[instrument_type]:
                        for instrument_group in new_instrument_list[instrument_type]:
                            subscription_packet = self.create_subscription_packet(instrument_group, int(instrument_type))
                            asyncio.ensure_future(self.ws.send(subscription_packet))
            elif self.version == 'v2':
                new_instrument_list = self.validate_and_process_tuples(symbols)
                for instrument_type, instrument_groups in new_instrument_list.items():
                    for instrument_group in instrument_groups:
                        for i in range(0, len(instrument_group), 100):
                            batch = instrument_group[i:i+100]
                            subscription_message = {
                                "RequestCode": int(instrument_type),
                                "InstrumentCount": len(batch),
                                "InstrumentList": [
                                    {
                                        "ExchangeSegment": self.get_exchange_segment(ex),
                                        "SecurityId": token
                                    } for ex, token in batch
                                ]
                            }
                            asyncio.ensure_future(self.ws.send(json.dumps(subscription_message)))

    def unsubscribe_symbols(self, symbols):
        """Function to unsubscribe symbols from connection when connection is already active."""
        # Update the instruments list by removing the specified symbols
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.difference_update(symbols)
        self.instruments = list(unique_symbols_set)

        # If the WebSocket is open, send the unsubscription packet for the symbols
        if self.ws and not self.ws.closed:
            # Prepare the instruments list for unsubscription
            group_size = 100
            instrument_list_to_unsubscribe = self.validate_and_process_tuples(symbols, group_size)

            if self.version == 'v1':
            # Send unsubscription packets for the removed symbols
                for instrument_type in instrument_list_to_unsubscribe:
                    if instrument_list_to_unsubscribe[instrument_type]:
                        for instrument_group in instrument_list_to_unsubscribe[instrument_type]:
                            # Use the original instrument_type for grouping, but increment for the actual request
                            unsubscription_packet = self.create_subscription_packet(instrument_group, int(instrument_type) + 1)
                            asyncio.ensure_future(self.ws.send(unsubscription_packet))

            elif self.version == 'v2':
                for instrument_type, instrument_groups in instrument_list_to_unsubscribe.items():
                    for instrument_group in instrument_groups:
                        for i in range(0, len(instrument_group), 100):
                            batch = instrument_group[i:i+100]
                            unsubscription_message = {
                                # Use the incremented instrument_type for the actual request
                                "RequestCode": int(instrument_type) + 1,
                                "InstrumentCount": len(batch),
                                "InstrumentList": [
                                    {
                                        "ExchangeSegment": self.get_exchange_segment(ex),
                                        "SecurityId": token
                                    } for ex, token in batch
                                ]
                            }
                            asyncio.ensure_future(self.ws.send(json.dumps(unsubscription_message)))


from .dhan_context import DhanContext
from .dhan_http import DhanHTTP
from ._order import Order
from ._forever_order import ForeverOrder
from ._portfolio import Portfolio
from ._funds import Funds
from ._statement import Statement
from ._trader_control import TraderControl
from ._security import Security
from ._market_feed import MarketFeed
from ._historical_data import HistoricalData
from ._option_chain import OptionChain
from .dhanhq import dhanhq
from .marketfeed import MarketFeed
from .orderupdate import OrderUpdate
from .fulldepth import FullDepth